/*
   ps_escape.c - Function to add escape sequences for PostScript strings,
                 for deckmate program.jj

   Author: Paul Hardy, August 2020

   Copyright (C) 2020 Paul Hardy

   LICENSE:

      This program is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "deckmate.h"


/*
   Add escape sequences for PostScript show operator.
*/
void
ps_escape (char *inbuffer, int chars_per_line, int fortran_cc,
           char *safe_buffer, int *symbol, unsigned *fonts_used) {

   int  i, j, k;     /* loop variables      */
   int  output_col;   /* current column in output line */


   safe_buffer[0] = '\0';

   /* Skip over the carriage control format character if enabled */
   i = 0;
   if (fortran_cc == 1 || inbuffer[0] == 0xB || inbuffer[0] == 0xC) i = 1;
   j = 0;
   /* Line is started.  Process one character at a time. */
   output_col = 1;
   while (output_col  <= chars_per_line              &&
          inbuffer[i] != '\r' && inbuffer[i] != '\n' &&
          inbuffer[i] != '\0') {

      /* Output spaces for one or more tabs and/or spaces in sequence */
      while (inbuffer[i] == '\t' && output_col <= chars_per_line) {

         k = 8 - ((output_col - 1) % 8);  /* spaces to add */
         output_col += k;
         while (k-- > 0) safe_buffer[j++] = ' ';

         i++;
      }

      /* Escape some characters for well-formed PostScript strings. */
      if (inbuffer[i] == '(' || inbuffer[i] == ')' ||
          inbuffer[i] == '/' || inbuffer[i] == '\\') {
         safe_buffer[j++] = '\\';
         safe_buffer[j++] = inbuffer[i];
         output_col++;
         i++;
      }
      else if (inbuffer[i] < 0x20 || inbuffer[i] >= 0x7F) {
         /*
            Check for "cents" symbol (U+00A2), as a single
            binary byte value or as UTF-8.
         */
         if ((inbuffer[i] == '\242') ||
             (inbuffer[i] == '\302' && inbuffer[i+1] == '\242')) {
            strcpy (&safe_buffer[j], "\\242");  /* cents symbol */
            j += 4;
            output_col++;
            if (inbuffer[i] == '\302') i++;  /* if two-byte UTF-8 */
            i++;
         }
         /*
            Check for "logical not" symbol (U+00AC), as a single
            binary byte value or as UTF-8.
         */
         else if ((inbuffer[i] == '\254') ||
                  (inbuffer[i] == '\302' && inbuffer[i+1] == '\254')) {
            /*
               Logical not symbol; flag as a PostScript font dagger
               glyph for ps_line_out function to replace with the
               logical not glyph from PostScript Symbol font.
            */
            /* '\330' will be printed as the Symbol logical not glyph */
            strcpy (&safe_buffer[j], "\\330");
            j += 4;
            output_col++;
            if (inbuffer[i] == '\302') i++;  /* if two-byte UTF-8 */
            i++;
            *symbol = 1;  /* mark output string as containing this glyph */
            *fonts_used |= FONT_PS_SYMBOL;
         }
         else {
            strcpy (&safe_buffer[j], "\\263");
            j += 4;
            output_col++;
            /*
               If this looks like a valid UTF-8 code point in the range
               U+0080..U+00FF (\302\200..\303\277), treat this input
               byte and the one that follows as one character.
            */
            if ((inbuffer[i] == '\302' || inbuffer[i] == '\303') &&
                ((unsigned)inbuffer[i+1] >= (unsigned)'\200' &&
                 (unsigned)inbuffer[i+1] <= (unsigned)'\277')) {
                  i++;
            }
            i++;
         }  /* Unknown or unprintable character; print a double dagger. */
      }  /* Byte is in range U+0000..U+001F or U+0080..U+00FF */
      else {
         safe_buffer[j++] = inbuffer[i];
         output_col++;
         i++;
      }
   }
   safe_buffer[j] = '\0';


   return;
}

